# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""
Various utilities
"""
from hashlib import md5
import traceback
import random
import string
import simplejson as json
import re
import os
import logging
from ConfigParser import NoOptionError


random.seed()
_RE_CODE = re.compile('[A-Z0-9]{4}-[A-Z0-9]{4}-[A-Z0-9]{4}-[A-Z0-9]{4}')


def randchar(chars=string.digits + string.letters):
    """Generates a random char using urandom.

    If the system does not support it, the function fallbacks on random.choice

    See Haypo's explanation on the used formula to pick a char:
    http://bitbucket.org/haypo/hasard/src/tip/doc/common_errors.rst
    """
    try:
        pos = int(float(ord(os.urandom(1))) * 256. / 255.)
        return chars[pos % len(chars)]
    except NotImplementedError:
        return random.choice(chars)


def _resolve_name(name):
    """Resolves the name and returns the corresponding object."""
    ret = None
    parts = name.split('.')
    cursor = len(parts)
    module_name = parts[:cursor]
    last_exc = None

    while cursor > 0:
        try:
            ret = __import__('.'.join(module_name))
            break
        except ImportError, exc:
            last_exc = exc
            if cursor == 0:
                raise
            cursor -= 1
            module_name = parts[:cursor]

    for part in parts[1:]:
        try:
            ret = getattr(ret, part)
        except AttributeError:
            if last_exc is not None:
                raise last_exc
            raise ImportError(name)

    if ret is None:
        if last_exc is not None:
            raise last_exc
        raise ImportError(name)

    return ret


class CatchErrorMiddleware(object):
    """Middleware that catches error, log them and return a 500"""
    def __init__(self, app, config):
        self.app = app
        try:
            logger_name = config.get('global', 'logger_name')
        except NoOptionError:
            logger_name = 'root'

        self.logger = logging.getLogger(logger_name)
        try:
            hook = config.get('global', 'logger_hook')
            self.hook = _resolve_name(hook)
        except NoOptionError:
            self.hook = None

        try:
            self.ctype = config.get('global', 'logger_type')
        except NoOptionError:
            self.ctype = 'application/json'

    def __call__(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except:
            err = traceback.format_exc()
            hash = create_hash(err)
            self.logger.error(hash)
            self.logger.error(err)
            start_response('500 Internal Server Error',
                           [('content-type', self.ctype)])

            response = json.dumps("application error: crash id %s" % hash)
            if self.hook:
                try:
                    response = self.hook({'error': err, 'crash_id': hash,
                                          'environ': environ})
                except Exception:
                    pass

            return [response]


def make_err_mdw(app, global_conf, **conf):
    config = app.registry.settings['config']
    return CatchErrorMiddleware(app, config)


def create_hash(data):
    """Creates a unique hash using the data provided
    and a bit of randomness
    """
    rand = ''.join([randchar() for x in range(10)])
    data += rand
    return md5(data + rand).hexdigest()



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import time
import unittest2
import tempfile

from zope.interface import implements

import pyramid.testing
import pyramid.request
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import authenticated_userid
from pyramid.httpexceptions import HTTPUnauthorized

import tokenlib
import macauthlib

from mozsvc.exceptions import BackendError
from mozsvc.tests.support import TestCase

try:
    from mozsvc.user.noncecache import MemcachedNonceCache
    MEMCACHED = True
except ImportError:
    MEMCACHED = False


DEFAULT_SETTINGS = {
    'cef.vendor': 'mozilla',
    'cef.device_version': '1.3',
    'cef.product': 'weave',
    'cef.use': True,
    'cef.version': 0,
   'cef.file': 'syslog',
}


class ExpandoRequest(object):
    """Proxy class for setting arbitrary attributes on a request.

    This class allows us to easily stub out various derived properties of
    a Request object.  Mostly notable, Request.host_url which is usually
    a read-only property derviced from the request evnironment.
    """

    def __init__(self, req):
        self.__dict__["_ExpandoRequest__req"] = req

    def __getattr__(self, attr):
        return getattr(self.__req, attr)

    def __setattr__(self, attr, value):
        try:
            setattr(self.__req, attr, value)
        except AttributeError:
            self.__dict__[attr] = value


class StubAuthenticationPolicy(object):
    """Authentication policy taking creds from request headers.

    This is a stub authentication policy that takes username from the
    X-Username header, checks that it matches the password provided in
    the X-Password header, and returns that as the userid.
    """

    implements(IAuthenticationPolicy)

    def authenticated_userid(self, request):
        username = request.environ.get("HTTP_X_USERNAME")
        if not username:
            return None
        password = request.environ.get("HTTP_X_PASSWORD")
        if not password:
            return None
        # Don't panic!  This is only used in the tests as an easy way
        # to simulate both failed and successfull logins.
        if password != username:
            return None
        request.user["x-was-ere"] = True
        return username

    def unauthenticated_userid(self, request):
        raise RuntimeError("tests shouldn't call this")  # pragma: nocover

    def effective_principals(self, request):
        raise RuntimeError("tests shouldn't call this")  # pragma: nocover


class UserTestCase(TestCase):

    def setUp(self):
        super(UserTestCase, self).setUp()
        self.config.commit()
        self.policy = self.config.registry.queryUtility(IAuthenticationPolicy)

    def get_configurator(self):
        config = super(UserTestCase, self).get_configurator()
        config.add_settings(DEFAULT_SETTINGS)
        config.include("mozsvc.user")
        return config

    def test_that_authn_policy_can_be_overridden(self):
        self.config.set_authentication_policy(StubAuthenticationPolicy())
        self.config.commit()
        # Good password => successful auth via request.user
        request = self.make_request(environ={
                    "HTTP_X_USERNAME": "user1",
                    "HTTP_X_PASSWORD": "user1",
                  })
        self.assertEquals(request.user["uid"], "user1")
        self.assertEquals(request.user["x-was-ere"], True)
        # Bad password => request.user is empty
        request = self.make_request(environ={
                    "HTTP_X_USERNAME": "user1",
                    "HTTP_X_PASSWORD": "BAD PASSWORD",
                  })
        self.assertFalse(request.user)
        # No username => request.user is empty
        request = self.make_request(environ={
                    "HTTP_X_PASSWORD": "password1",
                  })
        self.assertFalse(request.user)
        # No password => request.user is empty
        request = self.make_request(environ={
                    "HTTP_X_USERNAME": "user1",
                  })
        self.assertFalse(request.user)

    def test_that_macauth_is_used_by_default(self):
        # Generate signed request.
        req = self.make_request()
        tokenid, key = self.policy.encode_mac_id(req, 42)
        macauthlib.sign_request(req, tokenid, key)
        # That should be enough to authenticate.
        self.assertEquals(authenticated_userid(req), 42)
        self.assertEquals(req.user.get("uid"), 42)
        # Check that it rejects invalid MAC ids.
        req = self.make_request()
        macauthlib.sign_request(req, tokenid, key)
        authz = req.environ["HTTP_AUTHORIZATION"]
        req.environ["HTTP_AUTHORIZATION"] = authz.replace(tokenid, "XXXXXX")
        self.assertRaises(HTTPUnauthorized, authenticated_userid, req)
        # And that the rejection gets raised when accessing request.user
        self.assertRaises(HTTPUnauthorized, getattr, req, "user")

    def test_that_req_user_can_be_replaced(self):
        req = self.make_request()
        tokenid, key = self.policy.encode_mac_id(req, 42)
        macauthlib.sign_request(req, tokenid, key)
        req.user = {"uid": 7}
        self.assertEquals(req.user, {"uid": 7})

    def test_that_macauth_cant_use_both_secret_and_secrets_file(self):
        config2 = pyramid.testing.setUp()
        config2.add_settings(DEFAULT_SETTINGS)
        config2.add_settings({
            "macauth.secret": "DARTH VADER IS LUKE'S FATHER",
            "macauth.secrets_file": "/dev/null",
        })
        self.assertRaises(ValueError, config2.include, "mozsvc.user")

    def test_that_macauth_can_use_per_node_hostname_secrets(self):
        with tempfile.NamedTemporaryFile() as sf:
            # Write some secrets to a file.
            sf.write("http://host1.com,0001:secret11,0002:secret12\n")
            sf.write("https://host2.com,0001:secret21,0002:secret22\n")
            sf.write("https://host3.com:444,0001:secret31,0002:secret32\n")
            sf.flush()
            # Configure the plugin to load them.
            config2 = pyramid.testing.setUp()
            config2.add_settings(DEFAULT_SETTINGS)
            config2.add_settings({
                "macauth.secrets_file": sf.name,
            })
            config2.include("mozsvc.user")
            config2.commit()
            # It should accept a request signed with the old secret on host1.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host1.com",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret11")
            key = tokenlib.get_token_secret(id, secret="secret11")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should accept a request signed with the new secret on host1.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host1.com",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret12")
            key = tokenlib.get_token_secret(id, secret="secret12")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should reject a request signed with secret from other host.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host2.com",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret12")
            key = tokenlib.get_token_secret(id, secret="secret12")
            macauthlib.sign_request(req, id, key)
            self.assertRaises(HTTPUnauthorized, authenticated_userid, req)
            # It should reject a request over plain http when host2 is ssl.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host2.com",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret22")
            key = tokenlib.get_token_secret(id, secret="secret22")
            macauthlib.sign_request(req, id, key)
            self.assertRaises(HTTPUnauthorized, authenticated_userid, req)
            # It should accept a request signed with the new secret on host2.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host2.com",
                "wsgi.url_scheme": "https",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret22")
            key = tokenlib.get_token_secret(id, secret="secret22")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should accept a request to host1 with an explicit port number.
            # Use some trickery to give host_url a value with default port.
            req = ExpandoRequest(self.make_request(config=config2, environ={
                "HTTP_HOST": "host1.com:80",
                "wsgi.url_scheme": "http",
            }))
            req.host_url = "http://host1.com:80"
            id = tokenlib.make_token({"uid": 42}, secret="secret11")
            key = tokenlib.get_token_secret(id, secret="secret11")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should accept a request to host2 with an explicit port number.
            # Use some trickery to give host_url a value with default port.
            req = ExpandoRequest(self.make_request(config=config2, environ={
                "HTTP_HOST": "host2.com:443",
                "wsgi.url_scheme": "https",
            }))
            req.host_url = "https://host2.com:443"
            id = tokenlib.make_token({"uid": 42}, secret="secret22")
            key = tokenlib.get_token_secret(id, secret="secret22")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should accept a request to host3 on a custom port.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host3.com:444",
                "wsgi.url_scheme": "https",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret32")
            key = tokenlib.get_token_secret(id, secret="secret32")
            macauthlib.sign_request(req, id, key)
            self.assertEquals(authenticated_userid(req), 42)
            # It should reject unknown hostnames.
            req = self.make_request(config=config2, environ={
                "HTTP_HOST": "host4.com",
            })
            id = tokenlib.make_token({"uid": 42}, secret="secret12")
            key = tokenlib.get_token_secret(id, secret="secret12")
            macauthlib.sign_request(req, id, key)
            self.assertRaises(HTTPUnauthorized, authenticated_userid, req)


class TestMemcachedNonceCache(unittest2.TestCase):

    def setUp(self):
        self.nc = None

    def tearDown(self):
        if self.nc is not None:
            self.nc.mcclient.flush_all()

    def test_operation(self, now=lambda: int(time.time())):
        if not MEMCACHED:
            raise unittest2.SkipTest("no memcache")
        ttl = 1
        nc = self.nc = MemcachedNonceCache(nonce_ttl=ttl)
        # Initially nothing is cached, so all nonces as fresh.
        try:
            self.assertTrue(nc.check_nonce("id", now(), "abc"))
        except BackendError:
            raise unittest2.SkipTest("no memcache")
        # After adding a nonce, that nonce should no longer be fresh.
        self.assertFalse(nc.check_nonce("id", now(), "abc"))
        self.assertTrue(nc.check_nonce("id", now(), "xyz"))
        # After the ttl passes, the nonce should be expired.
        # Unfortunately memcached only supports integer ttls, so the
        # smallest amount of time we can sleep here is 1 second.
        time.sleep(ttl)
        self.assertTrue(nc.check_nonce("id", now(), "abc"))
        # If the timestamp is too old, even a fresh nonce will fail the check.
        self.assertFalse(nc.check_nonce("id", now() - 2 * ttl, "def"))
        self.assertFalse(nc.check_nonce("id", now() + 2 * ttl, "def"))
        self.assertTrue(nc.check_nonce("id", now(), "def"))

    def test_operation_with_backward_clock_skew(self):
        def now():
            return int(time.time()) - 13
        self.test_operation(now=now)

    def test_operation_with_forward_clock_skew(self):
        def now():
            return int(time.time()) + 7
        self.test_operation(now=now)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contributor(s):
#   Victor Ng (vng@mozilla.com)
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
"""
Utility functions to instantiate the metrics logger from the sync
server configuration, as well as a decorator to time arbitrary
functions.
"""

from contextlib import contextmanager
from cornice import Service
from metlog.config import client_from_dict_config
from metlog.decorators import timeit, incr_count
from metlog.decorators.base import MetlogDecorator
from metlog.holder import CLIENT_HOLDER
from metlog.logging import hook_logger
from metlog.senders.logging import StdLibLoggingSender
import threading

from mozsvc.plugin import load_from_settings


_LOCAL_STORAGE = threading.local()


# Default settings to use when there is no "metlog" section in the config file.
DEFAULT_METLOG_SETTINGS = {
    "metlog.backend": "mozsvc.metrics.MetlogPlugin",
    "metlog.sender_class": "metlog.senders.logging.StdLibLoggingSender",
    "metlog.sender_json_types": [],
}


def setup_metlog(config_dict, default=False):
    """
    Instantiate a Metlog client and add it to the client holder.

    :param config_dict: Dictionary object containing the metlog client
                        configuration.
    :param default: Should this be specified as CLIENT_HOLDER's default
                    client? Note that the first client to be added will
                    automatically be specified as the default, regardless
                    of the value of this argument.
    """
    name = config_dict.get('logger', '')
    client = CLIENT_HOLDER.get_client(name)
    client = client_from_dict_config(config_dict, client)
    if default:
        CLIENT_HOLDER.set_default_client_name(name)


def teardown_metlog():
    pass


def get_metlog_client(name=None):
    """
    Return the specified Metlog client from the CLIENT_HOLDER.

    :param name: Name of metlog client to fetch. If not provided the
                 holder's specified default client will be used.
    """
    if name is not None:
        client = CLIENT_HOLDER.get_client(name)
    else:
        client = CLIENT_HOLDER.default_client
    return client


class MetlogPlugin(object):
    def __init__(self, **kwargs):
        setup_metlog(kwargs)
        self.client = CLIENT_HOLDER.default_client


def load_metlog_client(config):
    """Load and return a metlog client for the given Pyramid Configurator.

    This is a shortcut function to load and return the metlog client specified
    by the given Pyramid Configurator object.  If the configuration does not
    specify any metlog settings, a default client is constructed that routes
    all messages into the stdlib logging routines.

    The metlog client is cached in the Configurator's registry, so multiple
    calls to this function will return a single instance.
    """
    settings = config.registry.settings
    client = config.registry.get("metlog")
    if client is None:
        if "metlog.backend" not in settings:
            settings.update(DEFAULT_METLOG_SETTINGS)
        client = load_from_settings('metlog', settings).client
        config.registry['metlog'] = client
        if not isinstance(client.sender, StdLibLoggingSender):
            hook_logger("mozsvc", client)
    return client


def get_tlocal():
    """
    Return the thread local metlog context dict, if it exists. This should only
    succeed from within a `thread_context` context manager. If we're not within
    such a context manager (and thus the metlog context dict doesn't exist)
    then an AttributeError will be raised.
    """
    if not hasattr(_LOCAL_STORAGE, 'metlog_context_dict'):
        raise AttributeError("No `metlog_context_dict`; are you in a "
                             "thread_context?")
    return _LOCAL_STORAGE.metlog_context_dict


@contextmanager
def thread_context(callback):
    """
    This is a context manager that accepts a callback function and returns a
    thread local dictionary object. Upon exit, the callback function will be
    called and passed that dictionary as the sole argument, after which the
    dictionary will be deleted.
    """
    _LOCAL_STORAGE.metlog_context_dict = dict()
    yield _LOCAL_STORAGE.metlog_context_dict
    try:
        callback(_LOCAL_STORAGE.metlog_context_dict)
    finally:
        del _LOCAL_STORAGE.metlog_context_dict


class send_mozsvc_data(MetlogDecorator):
    """
    Decorator that can be wrapped around a view method which will check the
    `metlog_context_dict` threadlocal and, if not empty, generate a metlog
    message of type `mozsvc` with the contained data.
    """
    def metlog_call(self, *args, **kwargs):
        def send_logmsg(mozsvc_data):
            """
            Stuff the threadlocal data into the message and send it out.
            """
            if mozsvc_data:
                self.client.metlog('mozsvc', fields=mozsvc_data)

        with thread_context(send_logmsg):
            return self._fn(*args, **kwargs)


def update_mozsvc_data(update_data):
    """
    Update the `metlog_context_dict` with data that will be sent out via metlog
    after request processing.
    """
    get_tlocal().update(update_data)


class MetricsService(Service):

    def __init__(self, **kw):
        self._decorators = kw.pop('decorators', [timeit, incr_count,
                                                 send_mozsvc_data])
        Service.__init__(self, **kw)

    def get_view_wrapper(self, kw):
        """
        Returns a wrapper that will wrap the view callable w/ metlog decorators
        for timing and logging wsgi variables.
        """
        decorators = kw.pop('decorators', self._decorators)

        def wrapper(func):
            applied_set = set()
            if hasattr(func, '_metlog_decorators'):
                applied_set.update(func._metlog_decorators)
            for decorator in decorators:
                # Stacked api decorators may result in this being called more
                # than once for the same function, we need to make sure that
                # the original function isn't wrapped more than once by the
                # same decorator.
                if decorator not in applied_set:
                    func = decorator(func)
                    applied_set.add(decorator)
            func._metlog_decorators = applied_set
            return func
        return wrapper



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Simplified API for accessing memcached.

This module builds upon the pylibmc.Client class to provide a slightly
higher-level interface to memcached.  It offers useful default behaviours
for serialization, error reporting and connection pooling.
"""

import sys
import time
import json
import traceback
import contextlib
import Queue

import pylibmc

from pyramid.threadlocal import get_current_registry

from mozsvc.exceptions import BackendError


class MemcachedClient(object):
    """Helper class for interfacing with pylibmc.

    This class provides the basic methods of the pylibmc.Client class, but
    wraps them with some extra functionality:

        * all values are transparently serialized via JSON instead of pickle.
        * connections are taken from an underlying pool.
        * errors are converted into BackendError instances.
        * cas() transparently falls back to add() when appropriate.

    """

    def __init__(self, servers=None, key_prefix="", pool_size=None,
                 pool_timeout=60, **kwds):
        if servers is None:
            servers = ["127.0.0.1:11211"]
        elif isinstance(servers, basestring):
            servers = [servers]
        self.key_prefix = key_prefix
        master = pylibmc.Client(servers, behaviors={"cas": 1})
        self.pool = MCClientPool(master, pool_size, pool_timeout)

    @property
    def logger(self):
        return get_current_registry()["metlog"]

    @contextlib.contextmanager
    def _connect(self):
        """Context mananager for getting a connection to memcached."""
        with self.pool.reserve() as mc:
            try:
                yield mc
            except pylibmc.Error, err:
                err = traceback.format_exc()
                self.logger.error(err)
                raise BackendError(str(err))

    def get(self, key):
        """Get the value stored under the given key."""
        with self._connect() as mc:
            data = mc.get(self.key_prefix + key)
        if data is not None:
            data = json.loads(data)
        return data

    def gets(self, key):
        """Get the current value and casid for the given key."""
        with self._connect() as mc:
            data, casid = mc.gets(self.key_prefix + key)
        if data is not None:
            data = json.loads(data)
        return data, casid

    def get_multi(self, keys):
        """Get the values stored under the given keys in a single request."""
        with self._connect() as mc:
            myitems = mc.get_multi([self.key_prefix + key for key in keys])
        items = {}
        for key, value in myitems.iteritems():
            assert key.startswith(self.key_prefix)
            items[key[len(self.key_prefix):]] = json.loads(value)
        return items

    def set(self, key, value, **kwds):
        """Set the value stored under the given key."""
        data = json.dumps(value)
        with self._connect() as mc:
            res = mc.set(self.key_prefix + key, data, **kwds)
        return res

    def add(self, key, value, **kwds):
        """Add the given key to memcached if not already present."""
        data = json.dumps(value)
        with self._connect() as mc:
            res = mc.add(self.key_prefix + key, data, **kwds)
        return res

    def cas(self, key, value, casid, **kwds):
        """Set the value stored under the given key if casid matches."""
        data = json.dumps(value)
        with self._connect() as mc:
            # Memcached's CAS only works properly on existing keys.
            # Fortunately ADD has the same semantics for missing keys.
            if casid is None:
                res = mc.add(self.key_prefix + key, data, **kwds)
            else:
                res = mc.cas(self.key_prefix + key, data, casid, **kwds)
        return res

    def delete(self, key):
        """Delete the value stored under the given key."""
        with self._connect() as mc:
            try:
                res = mc.delete(self.key_prefix + key)
            except pylibmc.NotFound:
                res = False
        return res

    def flush_all(self):
        """Delete all keys from memcached.  Obviously very dangerous..."""
        with self._connect() as mc:
            return mc.flush_all()


# Sentinel used to mark an empty slot in the MCClientPool queue.
# Using sys.maxint as the timestamp ensures that empty slots will always
# sort *after* live connection objects in the queue.
EMPTY_SLOT = (sys.maxint, None)


class MCClientPool(object):
    """Pool of pylibmc.Client objects, with periodic purging of connections.

    This class implements roughly the same interface as pylibmc.ClientPool,
    but it includes functionality to periodically close and refresh the
    pooled Client objects.  This seems to work around some occasional hangs
    that were occurring with long-lived clients.

    To initialise the pool you must provide a master Client object from which
    pool entries will be cloned.  You may also specify the maximum size of the
    pool and the time after which old connections will be recycled.

    To obtain a Client object from the pool, call reserve() as a context
    manager like this::

        with pool.reserve() as mc:
            mc.set("hello", "world")
            assert ms.get("hello") == "world"

    """

    def __init__(self, master, maxsize=None, timeout=60):
        self.master = master
        self.maxsize = maxsize
        self.timeout = timeout
        # Use a synchronized Queue class to hold the active client objects.
        # It will contain tuples (connection_timestamp, client).
        # Using a PriorityQueue ensures that the oldest connection is always
        # used first, allowing them to be closed out when stale.  It also means
        # that a no-maxsize pool can grow and shink according to demand, as old
        # connections are expired and not replaced.
        self.clients = Queue.PriorityQueue(maxsize)
        # If there is a maxsize, prime the queue with empty slots.
        if maxsize is not None:
            for _ in xrange(maxsize):
                self.clients.put(EMPTY_SLOT)

    @contextlib.contextmanager
    def reserve(self):
        """Context-manager to obtain a Client object from the pool."""
        ts, client = self._checkout_connection()
        try:
            yield client
        finally:
            self._checkin_connection(ts, client)

    def _checkout_connection(self):
        """Checkout a connection from the pool.

        This method checks out a connection from the pool, creating a new one
        if necessary.  It will block if a maxsize has been set and there are
        no connections left in the pool
        """
        # If there's no maxsize, no need to block waiting for a connection.
        blocking = (self.maxsize is not None)
        # Loop until we get a non-stale connection, or we create a new one.
        while True:
            try:
                ts, client = self.clients.get(blocking)
            except Queue.Empty:
                # No maxsize and no free connections, create a new one.
                # XXX TODO: we should be using a monotonic clock here.
                now = int(time.time())
                return now, self.master.clone()
            else:
                now = int(time.time())
                # If we got an empty slot placeholder, create a new connection.
                if client is None:
                    return now, self.master.clone()
                # If the connection is not stale, go ahead and use it.
                if ts + self.timeout > now:
                    return ts, client
                # Otherwise, the connection is stale.
                # Close it, push an empty slot onto the queue, and retry.
                client.disconnect_all()
                self.clients.put(EMPTY_SLOT)
                continue

    def _checkin_connection(self, ts, client):
        """Return a connection to the pool."""
        # If the connection is now stale, don't return it to the pool.
        # Push an empty slot instead so that it will be refreshed when needed.
        now = int(time.time())
        if ts + self.timeout > now:
            self.clients.put((ts, client))
        else:
            if self.maxsize is not None:
                self.clients.put(EMPTY_SLOT)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   Victor Ng (vng@mozilla.com)

# ***** END LICENSE BLOCK *****

from StringIO import StringIO
from mozsvc.config import Config
from mozsvc.plugin import load_and_register
from pyramid.config import Configurator
from textwrap import dedent
import unittest2
import json

metlog = True
try:
    from metlog.decorators import timeit, incr_count
    from metlog.holder import CLIENT_HOLDER
except ImportError:
    metlog = False


class TestDisabledTimers(unittest2.TestCase):
    """
    We want the counter decorators to fire, but the timer decorators should not
    """
    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))
        self.orig_globals = CLIENT_HOLDER.global_config.copy()
        config = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        logger = test
        sender_class=metlog.senders.DebugCaptureSender
        global_disabled_decorators = timeit
                                     something
        """)))
        settings = {"config": config}
        config = Configurator(settings=settings)
        self.plugin = load_and_register("test1", config)
        config.commit()

    def tearDown(self):
        CLIENT_HOLDER.delete_client('test')
        CLIENT_HOLDER.global_config = self.orig_globals

    def test_only_some_decorators(self):
        plugin = self.plugin

        plugin.client.sender.msgs.clear()
        self.assertEqual(len(plugin.client.sender.msgs), 0)

        @incr_count
        @timeit
        def no_timer(x, y):
            return x + y

        no_timer(5, 6)
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 1)

        expected = 'mozsvc.tests.test_partial_disable_metrics.no_timer'
        self.assertEqual(msgs[0]['fields']['name'], expected)
        self.assertEqual(msgs[0]['type'], 'counter')

        plugin.client.sender.msgs.clear()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Helper APIs for accessing various storage services.

This sub-package provides helper APIs for accessing various backend storage
services.  Currently available are:

    * mozsvc.storage.mcclient:  client for interacting with memcache

More may be added in the future, e.g. an SQL database access layer.
"""
